import copy
import logging

from dramatiq.middleware import Middleware

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import (
    PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL,
    DEFAULT_RESULT_TTL,
)
from pypeline.utils.redis_client import ResilientRedisClient

logger = logging.getLogger(__name__)


class ParallelPipeline(Middleware):
    def __init__(self, redis_url):
        self.redis_url = redis_url
        self.redis_conn = ResilientRedisClient(redis_url)

    def _get_execution_id(self, message):
        return message.kwargs.get("event", {}).get("execution_id")

    def _get_pipeline_id(self, message):
        return message.kwargs.get("event", {}).get("pipeline_id")

    def _get_pipeline_name(self, message):
        return message.kwargs.get("event", {}).get("pipeline_name")

    def _get_dispatching_key(self, execution_id: str, group_index: int) -> str:
        return f"pp:{execution_id}:group:{group_index}:dispatching"

    def _get_dispatched_key(self, execution_id: str, group_index: int) -> str:
        return f"pp:{execution_id}:group:{group_index}:dispatched"

    def after_process_message(self, broker, message, *, result=None, exception=None):
        from dramatiq.message import Message

        execution_id = self._get_execution_id(message)
        pipeline_id = self._get_pipeline_id(message)
        pipeline_name = self._get_pipeline_name(message)
        actor_name = getattr(message, "actor_name", None)
        message_id = getattr(message, "message_id", None)

        logger.info(
            "parallel_pipeline.after_process_message.start "
            "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s exception_present=%s",
            pipeline_id,
            pipeline_name,
            execution_id,
            actor_name,
            message_id,
            exception is not None,
        )

        try:
            if exception is not None:
                logger.info(
                    "parallel_pipeline.after_process_message.skip_exception "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                )
                return

            group_completion_uuid = message.options.get("group_completion_uuid")
            if not group_completion_uuid:
                logger.info(
                    "parallel_pipeline.after_process_message.skip_no_group_completion_uuid "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                )
                return

            execution_graph = message.options.get("execution_graph")
            if not execution_graph or not execution_id:
                logger.info(
                    "parallel_pipeline.after_process_message.skip_missing_execution_graph_or_execution_id "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s has_execution_graph=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    bool(execution_graph),
                )
                return

            locking_parallel_barrier = LockingParallelBarrier(
                self.redis_url,
                task_key=group_completion_uuid,
                lock_key=f"{group_completion_uuid}-lock",
            )

            try:
                logger.info(
                    "parallel_pipeline.barrier.acquire "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "group_completion_uuid=%s lock_key=%s timeout=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    group_completion_uuid,
                    f"{group_completion_uuid}-lock",
                    PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL,
                )
                locking_parallel_barrier.acquire_lock(
                    timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                )
                remaining_tasks = locking_parallel_barrier.decrement_task_count()
                logger.info(
                    "parallel_pipeline.barrier.decremented "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "group_completion_uuid=%s remaining_tasks=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    group_completion_uuid,
                    remaining_tasks,
                )
            finally:
                released = locking_parallel_barrier.release_lock()
                logger.info(
                    "parallel_pipeline.barrier.released "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "group_completion_uuid=%s released=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    group_completion_uuid,
                    released,
                )

            if remaining_tasks > 0:
                logger.info(
                    "parallel_pipeline.after_process_message.waiting_on_group "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "group_completion_uuid=%s remaining_tasks=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    group_completion_uuid,
                    remaining_tasks,
                )
                return

            for i in range(len(execution_graph)):
                message_group = execution_graph[i]

                if (
                    message_group[0]["options"]["group_completion_uuid"]
                    != group_completion_uuid
                ):
                    continue

                logger.info(
                    "parallel_pipeline.group_completed "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "completed_group_index=%s group_completion_uuid=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    i,
                    group_completion_uuid,
                )

                if i + 1 >= len(execution_graph):
                    logger.info(
                        "parallel_pipeline.pipeline_complete "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "completed_group_index=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        i,
                    )
                    return

                next_group_index = i + 1
                next_group = execution_graph[next_group_index]

                dispatching_key = self._get_dispatching_key(
                    execution_id, next_group_index
                )
                dispatched_key = self._get_dispatched_key(
                    execution_id, next_group_index
                )

                if self.redis_conn.safe_exists(dispatched_key):
                    logger.info(
                        "parallel_pipeline.next_group_already_dispatched "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "next_group_index=%s dispatched_key=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        next_group_index,
                        dispatched_key,
                    )
                    return

                won_dispatch = self.redis_conn.safe_set(
                    dispatching_key,
                    "1",
                    nx=True,
                    ex=60,
                )
                if not won_dispatch:
                    logger.info(
                        "parallel_pipeline.next_group_dispatch_lock_not_acquired "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "next_group_index=%s dispatching_key=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        next_group_index,
                        dispatching_key,
                    )
                    return

                logger.info(
                    "parallel_pipeline.next_group_dispatch_lock_acquired "
                    "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                    "next_group_index=%s dispatching_key=%s next_group_size=%s",
                    pipeline_id,
                    pipeline_name,
                    execution_id,
                    actor_name,
                    message_id,
                    next_group_index,
                    dispatching_key,
                    len(next_group),
                )

                try:
                    if self.redis_conn.safe_exists(dispatched_key):
                        logger.info(
                            "parallel_pipeline.next_group_already_dispatched_after_lock "
                            "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                            "next_group_index=%s dispatched_key=%s",
                            pipeline_id,
                            pipeline_name,
                            execution_id,
                            actor_name,
                            message_id,
                            next_group_index,
                            dispatched_key,
                        )
                        return

                    completion_uuid = next_group[0]["options"]["group_completion_uuid"]
                    next_barrier = LockingParallelBarrier(
                        self.redis_url,
                        task_key=completion_uuid,
                        lock_key=f"{completion_uuid}-lock",
                    )
                    next_barrier.set_task_count(len(next_group))

                    logger.info(
                        "parallel_pipeline.next_group_barrier_initialized "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "next_group_index=%s next_group_completion_uuid=%s next_group_size=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        next_group_index,
                        completion_uuid,
                        len(next_group),
                    )

                    execution_graph_copy = copy.deepcopy(execution_graph)

                    for idx, next_message in enumerate(next_group):
                        next_message["options"][
                            "execution_graph"
                        ] = execution_graph_copy
                        logger.info(
                            "parallel_pipeline.enqueue_next_message "
                            "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                            "next_group_index=%s next_message_index=%s next_actor_name=%s next_group_completion_uuid=%s",
                            pipeline_id,
                            pipeline_name,
                            execution_id,
                            actor_name,
                            message_id,
                            next_group_index,
                            idx,
                            next_message.get("actor_name"),
                            completion_uuid,
                        )
                        broker.enqueue(Message(**next_message))

                    self.redis_conn.safe_set(
                        dispatched_key,
                        "1",
                        ex=DEFAULT_RESULT_TTL,
                    )

                    logger.info(
                        "parallel_pipeline.next_group_dispatched "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "next_group_index=%s dispatched_key=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        next_group_index,
                        dispatched_key,
                    )
                finally:
                    self.redis_conn.safe_delete(dispatching_key)
                    logger.info(
                        "parallel_pipeline.next_group_dispatch_lock_released "
                        "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                        "next_group_index=%s dispatching_key=%s",
                        pipeline_id,
                        pipeline_name,
                        execution_id,
                        actor_name,
                        message_id,
                        next_group_index,
                        dispatching_key,
                    )

                return

            logger.warning(
                "parallel_pipeline.completed_group_not_found_in_execution_graph "
                "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                "group_completion_uuid=%s",
                pipeline_id,
                pipeline_name,
                execution_id,
                actor_name,
                message_id,
                group_completion_uuid,
            )

        except Exception:
            logger.exception(
                "parallel_pipeline.after_process_message.failed "
                "pipeline_id=%s pipeline_name=%s execution_id=%s actor_name=%s message_id=%s "
                "group_completion_uuid=%s",
                pipeline_id,
                pipeline_name,
                execution_id,
                actor_name,
                message_id,
                message.options.get("group_completion_uuid"),
            )
            raise



import json
import logging
import typing
from copy import copy
from uuid import uuid4

import networkx as nx
from dramatiq import get_broker

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import REDIS_URL, PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
from pypeline.utils.dramatiq_utils import register_lazy_actor
from pypeline.utils.module_utils import get_callable
from pypeline.utils.pipeline_utils import get_execution_graph

logger = logging.getLogger(__name__)


class Pypeline:
    def __init__(
        self,
        pipeline: dict,
        scenarios: typing.List[dict] = None,
        broker=None,
    ):
        if not scenarios:
            raise ValueError("At least one scenario is required")

        self.pipeline = pipeline
        self.broker = broker or get_broker()
        self._starting_messages = []
        self.scenarios = scenarios

        pipeline_config = pipeline["config"]
        self.graph = get_execution_graph(pipeline_config)
        self.number_of_tasks = len(self.graph.nodes)
        task_definitions = pipeline_config["taskDefinitions"]
        first_task = list(pipeline_config["dagAdjacency"].keys())[0]

        base_case_execution_id = None

        for scenario in self.scenarios:
            if self.scenarios.index(scenario) == 0:
                base_case_execution_id = scenario.get("execution_id", None) or str(
                    uuid4()
                )
                scenario["execution_id"] = base_case_execution_id
                scenario["base_case_execution_id"] = base_case_execution_id
                scenario["branched_off_execution_id"] = (
                    scenario.get("branched_off_execution_id") or base_case_execution_id
                )
                scenario["tasksToRunInScenario"] = list(self.graph.nodes)
                continue
            tasks_in_reruns = scenario["taskReruns"]

            tasks_in_replacements = list(scenario["taskReplacements"].keys())

            distinct_scenario_tasks = list(set(tasks_in_reruns + tasks_in_replacements))
            tasks_to_be_rerun_in_scenario = distinct_scenario_tasks

            tasks_to_be_rerun_in_scenario = list(
                set(
                    task
                    for task in distinct_scenario_tasks
                    for task in nx.descendants(self.graph, task)
                )
                | set(tasks_to_be_rerun_in_scenario)
            )

            self.number_of_tasks = self.number_of_tasks + len(
                tasks_to_be_rerun_in_scenario
            )
            scenario["tasksToRunInScenario"] = tasks_to_be_rerun_in_scenario
            scenario["base_case_execution_id"] = base_case_execution_id
            scenario["execution_id"] = scenario.get("execution_id", None) or str(
                uuid4()
            )
            scenario["branched_off_execution_id"] = (
                scenario.get("branched_off_execution_id") or base_case_execution_id
            )

            if first_task in tasks_to_be_rerun_in_scenario:
                handler = task_definitions[first_task]["handlers"][
                    scenario["taskReplacements"].get(first_task, 0)
                ]
                server_type = task_definitions[first_task].get("serverType", None)
                lazy_actor = register_lazy_actor(
                    self.broker,
                    get_callable(handler),
                    pipeline_config["metadata"],
                    server_type,
                )
                message = lazy_actor.message()
                message.options["pipeline"] = pipeline
                if pipeline_config["metadata"].get("maxRetry", None) is not None:
                    message.options["max_retries"] = pipeline_config["metadata"][
                        "maxRetry"
                    ]
                message.options["task_replacements"] = copy(
                    scenario["taskReplacements"]
                )
                message.options["execution_id"] = scenario["execution_id"]
                message.options["task_name"] = first_task
                message.options["base_case_execution_id"] = base_case_execution_id
                message.options["branched_off_execution_id"] = scenario[
                    "branched_off_execution_id"
                ]
                if scenario["settings"]:
                    message.kwargs["settings"] = copy(scenario["settings"])
                    message.kwargs["settings"]["execution_id"] = scenario[
                        "execution_id"
                    ]
                    message.kwargs["settings"][
                        "base_case_execution_id"
                    ] = base_case_execution_id
                    message.kwargs["settings"]["branched_off_execution_id"] = scenario[
                        "branched_off_execution_id"
                    ]

                self._starting_messages.append(message)

        for m in self._starting_messages:
            m.options["scenarios"] = self.scenarios

        first_scenario_task_replacements = scenarios[0]["taskReplacements"]
        first_scenario_settings = scenarios[0].get("settings", None)

        handler = task_definitions[first_task]["handlers"][
            first_scenario_task_replacements.get(first_task, 0)
        ]
        server_type = task_definitions[first_task].get("serverType", None)
        lazy_actor = register_lazy_actor(
            self.broker,
            get_callable(handler),
            pipeline_config["metadata"],
            server_type,
        )
        message = lazy_actor.message()
        message.options["pipeline"] = pipeline
        if pipeline_config["metadata"].get("maxRetry", None) is not None:
            message.options["max_retries"] = pipeline_config["metadata"]["maxRetry"]
        message.options["task_replacements"] = first_scenario_task_replacements
        message.options["execution_id"] = base_case_execution_id
        message.options["task_name"] = first_task
        message.options["scenarios"] = self.scenarios
        message.options["base_case_execution_id"] = base_case_execution_id
        base_branched_off = (
            scenarios[0].get("branched_off_execution_id") or base_case_execution_id
        )
        message.options["branched_off_execution_id"] = base_branched_off

        if first_scenario_settings:
            message.kwargs["settings"] = copy(first_scenario_settings)
            message.kwargs["settings"]["execution_id"] = base_case_execution_id
            message.kwargs["settings"][
                "base_case_execution_id"
            ] = base_case_execution_id
            message.kwargs["settings"]["branched_off_execution_id"] = base_branched_off

        self._starting_messages.append(message)

    def run(self, *, delay=None):
        for message in self._starting_messages:
            task_key = (
                f"{message.options['execution_id']}-{message.options['task_name']}"
            )
            locking_parallel_barrier = LockingParallelBarrier(
                REDIS_URL,
                task_key=task_key,
                lock_key=f"{message.options['base_case_execution_id']}-lock",
            )
            locking_parallel_barrier.set_task_count(1)
            logger.info(
                "pypeline.run.enqueue " "execution_id=%s task_name=%s task_key=%s",
                message.options["execution_id"],
                message.options["task_name"],
                task_key,
            )
            self.broker.enqueue(message, delay=delay)

        return self

    def __len__(self):
        return self.number_of_tasks

    def completed(self):
        locks = []

        for scenario in self.scenarios:
            locks.append(
                {
                    "scenario_task_keys": [
                        f"{scenario['execution_id']}-{task}"
                        for task in scenario["tasksToRunInScenario"]
                    ],
                    "redis_lock_key": f"{scenario['base_case_execution_id']}-lock",
                }
            )

        for lock in locks:
            for task_key in lock["scenario_task_keys"]:
                locking_parallel_barrier = LockingParallelBarrier(
                    REDIS_URL, task_key=task_key, lock_key=lock["redis_lock_key"]
                )
                try:
                    locking_parallel_barrier.acquire_lock(
                        timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                    )
                    task_complete = True
                    if locking_parallel_barrier.task_exists():
                        remaining_tasks = locking_parallel_barrier.get_task_count()
                        if remaining_tasks >= 1:
                            task_complete = False
                            logger.info(
                                "pypeline.completed.task_pending task_key=%s remaining=%s",
                                task_key,
                                remaining_tasks,
                            )
                    else:
                        task_complete = False
                        logger.info(
                            "pypeline.completed.task_not_found task_key=%s",
                            task_key,
                        )
                finally:
                    locking_parallel_barrier.release_lock()
                if not task_complete:
                    return task_complete

        logger.info("pypeline.completed.all_tasks_done")
        return True

    def to_json(self) -> str:
        return json.dumps(
            {
                "pipeline": self.pipeline,
                "scenarios": self.scenarios,
            }
        )

    @classmethod
    def from_json(cls, json_data: str) -> typing.Type["Pypeline"]:
        data = json.loads(json_data)

        return cls(
            data["pipeline"],
            scenarios=data["scenarios"],
        )



from datetime import date

from marshmallow import Schema, fields, validate, ValidationError, validates_schema, INCLUDE


class MissingSettingsException(Exception):
    pass


def create_pipeline_settings_schema(pipeline_settings_schema_data):
    """
    Dynamically create a schema to validate user data based on settings.

    Args:
        pipeline_settings_schema_data (dict): The settings schema data containing
                                              field configurations.

    Returns:
        Schema: A dynamically created schema class for validating user data.
    """

    # Dictionary to store dynamically generated fields
    schema_fields = {}

    for key, config in pipeline_settings_schema_data["properties"].items():
        data_type = config.get("dataType")
        input_type = config.get("inputType")
        field_args = {}

        # Map dataType to Marshmallow field type
        field_type = {
            "string": fields.String,
            "int": fields.Integer,
            "float": fields.Float,
            "boolean": fields.Boolean,
            "datetime": fields.DateTime,
            "array": fields.List,
            "object": fields.Nested,
            "date": fields.Date,
        }.get(data_type)

        if not field_type:
            raise ValidationError(f"Unsupported dataType `{data_type}` for `{key}`.")

        # Handle array type
        if data_type == "array":
            element_type = config.get("elementType")
            if not element_type:
                raise ValidationError(f"`elementType` is required for array `{key}`.")
            field_args["cls_or_instance"] = {
                "string": fields.String,
                "int": fields.Integer,
                "float": fields.Float,
                "boolean": fields.Boolean,
                "datetime": fields.DateTime,
                "date": fields.Date,
            }.get(element_type)
            if not field_args["cls_or_instance"]:
                raise ValidationError(
                    f"Unsupported elementType `{element_type}` for array `{key}`."
                )

        # Handle object type
        if data_type == "object":
            properties = config.get("properties")
            if not properties:
                raise ValidationError(f"`properties` is required for object `{key}`.")
            # Recursively create a schema for the nested object
            nested_schema = create_pipeline_settings_schema({"properties": properties})
            field_args["schema"] = nested_schema

        # Handle range validation for numeric fields
        if data_type in ["int", "float"]:
            if "minimum" in config or "maximum" in config:
                field_args["validate"] = validate.Range(
                    min=config.get("minimum"), max=config.get("maximum")
                )

        # Handle dropdown or radio input options
        if input_type in ["dropdown", "radio"] and "options" in config:
            allowed_values = [option["value"] for option in config["options"]]
            field_args["validate"] = validate.OneOf(allowed_values)

        # Mark the field as required if specified
        if key in pipeline_settings_schema_data.get("required", []):
            field_args["required"] = True

        # Create the field and add to the schema fields dictionary
        schema_fields[key] = field_type(**field_args)

    # Dynamically create a schema class with the generated fields
    DynamicPipelineSettingsSchema = type(
        "DynamicPipelineSettingsSchema", (Schema,), schema_fields
    )

    return DynamicPipelineSettingsSchema()


def create_pipeline_settings_schema(pipeline_settings_schema_data):
    """
    Dynamically create a schema to validate user data based on settings.

    Args:
        pipeline_settings_schema_data (dict): The settings schema data containing
                                              field configurations.

    Returns:
        Schema: A dynamically created schema class for validating user data.
    """

    # Dictionary to store dynamically generated fields
    schema_fields = {}

    for key, config in pipeline_settings_schema_data["properties"].items():
        data_type = config.get("dataType")
        input_type = config.get("inputType")
        field_args = {}

        # Map dataType to Marshmallow field type
        field_type = {
            "string": fields.String,
            "int": fields.Integer,
            "float": fields.Float,
            "boolean": fields.Boolean,
            "datetime": fields.DateTime,
            "date": fields.Date,
            "array": fields.List,
            "object": fields.Nested,
        }.get(data_type)

        if not field_type:
            raise ValidationError(f"Unsupported dataType `{data_type}` for `{key}`.")

        # Handle array type
        if data_type == "array":
            element_type = config.get("elementType")
            if not element_type:
                raise ValidationError(f"`elementType` is required for array `{key}`.")
            field_args["cls_or_instance"] = {
                "string": fields.String,
                "int": fields.Integer,
                "float": fields.Float,
                "boolean": fields.Boolean,
                "datetime": fields.DateTime,
                "date": fields.Date,
            }.get(element_type)
            if not field_args["cls_or_instance"]:
                raise ValidationError(
                    f"Unsupported elementType `{element_type}` for array `{key}`."
                )

        # Handle object type
        if data_type == "object":
            properties = config.get("properties")
            if not properties:
                raise ValidationError(f"`properties` is required for object `{key}`.")
            # Recursively create a schema for the nested object
            nested_schema = create_pipeline_settings_schema({"properties": properties})
            # Use the nested schema as the `nested` argument for fields.Nested
            field_type = fields.Nested(nested_schema)

        # Handle range validation for numeric fields
        if data_type in ["int", "float"]:
            if "minimum" in config or "maximum" in config:
                field_args["validate"] = validate.Range(
                    min=config.get("minimum"), max=config.get("maximum")
                )

        # Handle dropdown or radio input options
        if input_type in ["dropdown", "radio"] and "options" in config:
            allowed_values = [option["value"] for option in config["options"]]
            field_args["validate"] = validate.OneOf(allowed_values)

        # Mark the field as required if specified
        if key in pipeline_settings_schema_data.get("required", []):
            field_args["required"] = True
        else:
            field_args["required"] = False
            field_args["allow_none"] = True

        # Create the field and add to the schema fields dictionary
        if data_type == "object":
            schema_fields[key] = field_type
        else:
            schema_fields[key] = field_type(**field_args)

    schema_fields["Meta"] = type(
        "Meta", (), {"unknown": INCLUDE}
    )
    # Dynamically create a schema class with the generated fields
    DynamicPipelineSettingsSchema = type(
        "DynamicPipelineSettingsSchema", (Schema,), schema_fields
    )

    return DynamicPipelineSettingsSchema()


class OptionSchema(Schema):
    label = fields.String(
        required=True,
        metadata={"description": "The display label for the option"},
    )
    value = fields.Raw(
        required=True,
        metadata={"description": "The value corresponding to the option"},
    )


def validate_min_max(data):
    """Custom validator to ensure min/max match the dataType."""
    data_type = data.get("dataType")
    minimum = data.get("minimum")
    maximum = data.get("maximum")

    if data_type in ["int", "float"]:
        if minimum is not None and not isinstance(
            minimum, (int if data_type == "int" else float)
        ):
            raise ValidationError(f"`minimum` must be of type {data_type}.")
        if maximum is not None and not isinstance(
            maximum, (int if data_type == "int" else float)
        ):
            raise ValidationError(f"`maximum` must be of type {data_type}.")
        if minimum is not None and maximum is not None and minimum > maximum:
            raise ValidationError("`minimum` must be less than or equal to `maximum`.")
    elif data_type not in ["int", "float"] and (
        minimum is not None or maximum is not None
    ):
        raise ValidationError(
            "`minimum` and `maximum` are only valid for numeric types (`int`, `float`)."
        )


class SettingSchema(Schema):
    dataType = fields.String(
        required=True,
        validate=validate.OneOf(
            ["string", "int", "float", "boolean", "datetime", "array", "object", "date"]
        ),
        metadata={"description": "The underlying data type of the setting"},
    )
    inputType = fields.String(
        required=True,
        validate=validate.OneOf(
            ["text", "dropdown", "radio", "checkbox", "searchable", "custom"]
        ),
        metadata={"description": "The type of input UI element"},
    )
    label = fields.String(
        required=True,
        metadata={"description": "The display label for the field"},
    )
    placeholder = fields.String(
        metadata={"description": "Placeholder text for text input fields"}
    )
    default = fields.Raw(
        required=False,
        metadata={"description": "Optional default value for the input field"}
    )
    minimum = fields.Raw(
        metadata={"description": "Minimum value for numeric data types"}
    )
    maximum = fields.Raw(
        metadata={"description": "Maximum value for numeric data types"}
    )
    options = fields.List(
        fields.Nested(OptionSchema),
        metadata={"description": "Options for dropdown or radio input types"},
    )
    searchEndpoint = fields.String(
        metadata={"description": "Endpoint for searchable fields"}
    )
    component = fields.String(
        metadata={"description": "React component for custom input types"}
    )
    elementType = fields.String(
        metadata={"description": "Element type for array data types"}
    )
    properties = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(lambda: SettingSchema),
        metadata={"description": "Properties for object data types"},
    )

    class Meta:
        ordered = True

    @validates_schema
    def validate_min_max(self, data, **kwargs):
        validate_min_max(data)

    @validates_schema
    def validate_options(self, data, **kwargs):
        """Ensure options are provided for dropdown or radio input types and validate value types."""
        input_type = data.get("inputType")
        options = data.get("options")
        data_type = data.get("dataType")

        if input_type in ["dropdown", "radio"]:
            if not options:
                raise ValidationError(
                    "`options` are required for dropdown and radio input types.",
                    field_name="options",
                )

            for option in options:
                value = option.get("value")
                if data_type == "int" and not isinstance(value, int):
                    raise ValidationError(
                        f"Option value `{value}` must be of type `int`."
                    )
                elif data_type == "float" and not isinstance(value, float):
                    raise ValidationError(
                        f"Option value `{value}` must be of type `float`."
                    )
                elif data_type == "boolean" and not isinstance(value, bool):
                    raise ValidationError(
                        f"Option value `{value}` must be of type `boolean`."
                    )
                elif data_type == "string" and not isinstance(value, str):
                    raise ValidationError(
                        f"Option value `{value}` must be of type `string`."
                    )
                elif data_type == "datetime" and not isinstance(
                    value, str
                ):  # Assuming ISO 8601 strings
                    raise ValidationError(
                        f"Option value `{value}` must be an ISO 8601 string for `datetime`."
                    )
                elif data_type == "date":
                    try:
                        date.fromisoformat(value)
                    except Exception:
                        raise ValidationError(
                            f"Option value `{value}` must be an ISO 8601 string for `date`."
                        )

    @validates_schema
    def validate_search_endpoint(self, data, **kwargs):
        """Ensure searchEndpoint is provided only for 'searchable' input types."""
        input_type = data.get("inputType")
        search_endpoint = data.get("searchEndpoint")

        if input_type == "searchable" and not search_endpoint:
            raise ValidationError(
                "`searchEndpoint` is required for `searchable` input types.",
                field_name="searchEndpoint",
            )
        elif input_type != "searchable" and search_endpoint:
            raise ValidationError(
                "`searchEndpoint` is not allowed for non-searchable input types.",
                field_name="searchEndpoint",
            )

    @validates_schema
    def validate_custom_component(self, data, **kwargs):
        """Ensure component is provided for custom input types."""
        input_type = data.get("inputType")
        component = data.get("component")

        if input_type == "custom" and not component:
            raise ValidationError(
                "`component` is required for `custom` input types.",
                field_name="component",
            )
        elif input_type != "custom" and component:
            raise ValidationError(
                "`component` is not allowed for non-custom input types.",
                field_name="component",
            )

    @validates_schema
    def validate_array_element_type(self, data, **kwargs):
        """Ensure elementType is provided for array data types."""
        data_type = data.get("dataType")
        element_type = data.get("elementType")

        if data_type == "array" and not element_type:
            raise ValidationError(
                "`elementType` is required for `array` data types.",
                field_name="elementType",
            )
        elif data_type != "array" and element_type:
            raise ValidationError(
                "`elementType` is not allowed for non-array data types.",
                field_name="elementType",
            )

    @validates_schema
    def validate_object_properties(self, data, **kwargs):
        """Ensure properties are provided for object data types."""
        data_type = data.get("dataType")
        properties = data.get("properties")

        if data_type == "object" and not properties:
            raise ValidationError(
                "`properties` is required for `object` data types.",
                field_name="properties",
            )
        elif data_type != "object" and properties:
            raise ValidationError(
                "`properties` is not allowed for non-object data types.",
                field_name="properties",
            )


class PipelineSettingsSchema(Schema):
    properties = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(SettingSchema),
        required=True,
        metadata={"description": "A dictionary of settings with their configurations"},
    )
    required = fields.List(
        fields.String(), required=True, description="List of required settings"
    )
    scenarioSettings = fields.List(
        fields.String(),
        required=False,
        description="List of settings that can be overriding for different pipeline scenarios.",
    )

    @validates_schema
    def validate_scenario_settings(self, data, **kwargs):
        """Ensure scenarioSettings only contains keys defined in properties."""
        properties = data.get("properties", {})
        scenario_settings = data.get("scenarioSettings", [])

        invalid_settings = [
            setting for setting in scenario_settings if setting not in properties
        ]
        if invalid_settings:
            raise ValidationError(
                {
                    "scenario_settings": (
                        f"The following settings in scenarioSettings are not defined "
                        f"in properties: {', '.join(invalid_settings)}"
                    )
                }
            )


class PipelineScenarioSchema(Schema):
    settings = fields.Dict(
        required=True,
        metadata={
            "description": "Settings to be used for a given scenario.  Should match the pypeline.yaml settings schema"
        },
    )
    taskReplacements = fields.Dict(
        keys=fields.String(),
        values=fields.Integer(),
        required=False,
        metadata={
            "description": "Tasks that should be replaced in a given scenario.  "
            "The key corresponds to the task definition in the pypeline.yaml and the value corresponds "
            "to the index of the task handlers where 0 is the default and first task.  Eg: {'a': 1}.  In this case "
            "if we have a task definition 'a' with 3 handlers fn_1, fn_2, fn_3 respectively then the handler to run "
            "for 'a' is fn_2."
        },
    )

    taskReruns = fields.List(
        fields.String(),
        required=False,
        metadata={
            "description": "List of task definitions that need to be run again for a given scenario.  Here "
            "the scenario's pipeline settings will be injected in the task being run again which could be used to "
            "produce alternative calculations and or results."
        },
    )
    execution_id = fields.String(required=False, metadata={"description":"Execution id for a known scenario"})
    branched_off_execution_id = fields.String(
        required=False,
        allow_none=True,
        metadata={"description": "Execution id of the scenario this one is branched from (or None for base/default)."},
    )


class PipelineScenariosSchema(Schema):
    required = fields.List(
        fields.Nested(PipelineScenarioSchema),
        metadata={"description": "List of scenarios to run for a given pipeline"},
    )


# Example usage
if __name__ == "__main__":
    pipeline_settings = {"param1": "test", "param2": 1}

    yaml_data = {
        "properties": {
            "param1": {
                "dataType": "string",
                "inputType": "text",
                "label": "Parameter 1",
                "placeholder": "Enter a string",
            },
            "param2": {
                "dataType": "int",
                "inputType": "text",
                "label": "Parameter 2",
                "minimum": -1,
                "maximum": 1,
            },
            "param3": {
                "dataType": "boolean",
                "inputType": "checkbox",
                "label": "Enable Feature",
            },
            "param4": {
                "dataType": "float",
                "inputType": "dropdown",
                "label": "Choose an Option",
                "minimum": 0.5,
                "maximum": 2.5,
                "options": [
                    {"label": "Option 1", "value": 0.5},
                    {"label": "Option 2", "value": 1.5},
                ],
            },
            "param5": {
                "dataType": "int",
                "inputType": "radio",
                "label": "Select a Mode",
                "options": [
                    {"label": "Mode A", "value": 1},
                    {"label": "Mode B", "value": 2},
                ],
            },
            "param6": {
                "dataType": "string",
                "inputType": "searchable",
                "label": "Select Pipeline",
                "searchEndpoint": "/api/pipelines",
            },
            "param7": {
                "dataType": "array",
                "inputType": "text",
                "label": "Array of Integers",
                "elementType": "int",
            },
            "param8": {
                "dataType": "object",
                "inputType": "custom",
                "label": "Custom Object",
                "component": "CustomObjectComponent",
                "properties": {
                    "subParam1": {
                        "dataType": "string",
                        "inputType": "text",
                        "label": "Sub Parameter 1",
                    },
                    "subParam2": {
                        "dataType": "int",
                        "inputType": "text",
                        "label": "Sub Parameter 2",
                    },
                },
            },
        },
        "required": ["param1", "param2", "param4"],
    }

    schema = PipelineSettingsSchema()
    errors = schema.validate(yaml_data)
    if errors:
        print("Validation errors:", errors)
    else:
        print("Validation successful!")



import logging
from typing import Union

from pypeline.constants import WORKER_NAME
from pypeline.pypeline_yaml import load_pypeline_config

logger = logging.getLogger(__name__)


def retrieve_latest_pipeline_config(
    pipeline_id: Union[str, None] = None
) -> Union[dict, list]:
    """Retrieve the 'latest' pipeline configuration for a given pipeline."""
    pypeline_config = load_pypeline_config()
    if "pipelines" in pypeline_config:
        pipelines = []
        found_pipeline = None
        for p_id, config in pypeline_config["pipelines"].items():
            if pipeline_id == p_id:
                found_pipeline = config
                break
            pipelines.append(config)

        if pipeline_id:
            if found_pipeline:
                return found_pipeline
            raise ValueError(f"Invalid pipeline {pipeline_id}")

        return pipelines
    return None


def retrieve_latest_schedule_config():
    """Retrieve the 'latest' scheduled tasks configuration."""
    pypeline_config = load_pypeline_config()
    if "scheduledTasks" in pypeline_config:
        tasks = []
        for task_id, config in pypeline_config["scheduledTasks"].items():
            tasks.append(config)
        return tasks
    return None


def retrieve_executable_job_config():
    pypeline_config = load_pypeline_config()

    if not pypeline_config:
        return None
    if "executableJobs" in pypeline_config:
        tasks = []
        for task_id, config in pypeline_config["executableJobs"].items():
            tasks.append(config)
        return tasks
    return None


def get_service_config_for_worker(
    pypeline_config: dict, worker_name: str = None
) -> Union[dict, None]:
    """For the current WORKER_NAME (which must be present in the environment
    of this worker instance for a valid deployment), return the worker's
    serviceConfig object.
    """
    if pypeline_config is None:
        raise ValueError("Pypeline config was not provided")
    if worker_name is None:
        worker_name = WORKER_NAME
    if worker_name is None:
        return None

    service_config = pypeline_config.get("serviceConfig", [])
    for service in service_config:
        if service["name"] == worker_name:
            return service

    raise ValueError(
        "Could not find a service config for worker "
        f"`{worker_name}`. Make sure you have added the service in"
        f" your pypeline.yaml with `name: {worker_name}` and "
        "`type: celery-worker`."
    )



import logging
from copy import copy

import networkx as nx
from dramatiq import Middleware

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import (
    PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL,
    DEFAULT_RESULT_TTL,
)
from pypeline.utils.redis_client import ResilientRedisClient
from pypeline.utils.module_utils import get_callable
from pypeline.utils.pipeline_utils import get_execution_graph
from pypeline.utils.dramatiq_utils import register_lazy_actor

logger = logging.getLogger(__name__)


class PypelineMiddleware(Middleware):

    def __init__(self, redis_url):
        self.redis_url = redis_url
        self.redis_conn = ResilientRedisClient(redis_url)

    def _child_dispatched_key(
        self, base_case_execution_id: str, execution_id: str, child: str
    ) -> str:
        return f"pypeline:{base_case_execution_id}:{execution_id}:{child}:dispatched"

    def after_process_message(self, broker, message, *, result=None, exception=None):

        pipeline_id = (
            message.kwargs.get("event", {}).get("pipeline_id")
            if message.kwargs.get("event")
            else None
        )
        pipeline_name = (
            message.kwargs.get("event", {}).get("pipeline_name")
            if message.kwargs.get("event")
            else None
        )
        actor_name = getattr(message, "actor_name", None)
        message_id = getattr(message, "message_id", None)

        logger.info(
            "pypeline.after_process_message.start "
            "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s exception_present=%s",
            pipeline_id,
            pipeline_name,
            actor_name,
            message_id,
            exception is not None,
        )

        if exception is not None:
            logger.info(
                "pypeline.after_process_message.skip_exception "
                "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s",
                pipeline_id,
                pipeline_name,
                actor_name,
                message_id,
            )
            return

        if "pipeline" not in message.options:
            logger.info(
                "pypeline.after_process_message.skip_no_pipeline "
                "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s",
                pipeline_id,
                pipeline_name,
                actor_name,
                message_id,
            )
            return

        pipeline = message.options["pipeline"]
        max_retries = message.options.get("max_retries", None)
        pipeline_config = pipeline["config"]
        task_replacements = message.options["task_replacements"]
        execution_id = message.options["execution_id"]
        base_case_execution_id = message.options["base_case_execution_id"]
        task_definitions = pipeline_config["taskDefinitions"]
        task_name = message.options["task_name"]
        task_key = f"{execution_id}-{task_name}"

        logger.info(
            "pypeline.after_process_message.processing "
            "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
            "execution_id=%s base_case_execution_id=%s task_name=%s",
            pipeline_id,
            pipeline_name,
            actor_name,
            message_id,
            execution_id,
            base_case_execution_id,
            task_name,
        )

        locking_parallel_barrier = LockingParallelBarrier(
            self.redis_url,
            task_key=task_key,
            lock_key=f"{base_case_execution_id}-lock",
        )
        try:
            locking_parallel_barrier.acquire_lock(
                timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
            )
            _ = locking_parallel_barrier.decrement_task_count()
        finally:
            locking_parallel_barrier.release_lock()

        logger.info(
            "pypeline.task_completed "
            "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
            "execution_id=%s task_name=%s task_key=%s",
            pipeline_id,
            pipeline_name,
            actor_name,
            message_id,
            execution_id,
            task_name,
            task_key,
        )

        graph = get_execution_graph(pipeline_config)
        children_tasks = pipeline_config["dagAdjacency"].get(task_name, [])
        messages = []
        for child in children_tasks:
            child_ancestors = sorted(graph.predecessors(child))
            child_ancestors_complete = {a: False for a in child_ancestors}

            tasks_to_run_in_scenario = None
            for scenario in message.options["scenarios"]:
                if scenario["execution_id"] == execution_id:
                    tasks_to_run_in_scenario = scenario["tasksToRunInScenario"]
                    break

            if tasks_to_run_in_scenario is None:
                logger.warning(
                    "pypeline.no_matching_scenario "
                    "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                    "execution_id=%s task_name=%s child=%s",
                    pipeline_id,
                    pipeline_name,
                    actor_name,
                    message_id,
                    execution_id,
                    task_name,
                    child,
                )
                continue

            for ancestor in child_ancestors:
                if ancestor in tasks_to_run_in_scenario:
                    current_scenario_ancestor_task_key = f"{execution_id}-{ancestor}"
                    locking_parallel_barrier = LockingParallelBarrier(
                        self.redis_url,
                        task_key=current_scenario_ancestor_task_key,
                        lock_key=f"{base_case_execution_id}-lock",
                    )
                    try:
                        locking_parallel_barrier.acquire_lock(
                            timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                        )
                        if not locking_parallel_barrier.task_exists():
                            child_ancestors_complete[ancestor] = False
                        elif locking_parallel_barrier.get_task_count() <= 0:
                            child_ancestors_complete[ancestor] = True
                    finally:
                        locking_parallel_barrier.release_lock()
                else:
                    base_scenario_ancestor_task_key = (
                        f"{base_case_execution_id}-{ancestor}"
                    )
                    locking_parallel_barrier = LockingParallelBarrier(
                        self.redis_url,
                        task_key=base_scenario_ancestor_task_key,
                        lock_key=f"{base_case_execution_id}-lock",
                    )
                    try:
                        locking_parallel_barrier.acquire_lock(
                            timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                        )
                        if not locking_parallel_barrier.task_exists():
                            child_ancestors_complete[ancestor] = False
                        elif locking_parallel_barrier.get_task_count() <= 0:
                            child_ancestors_complete[ancestor] = True
                    finally:
                        locking_parallel_barrier.release_lock()

            if any(complete is False for complete in child_ancestors_complete.values()):
                logger.info(
                    "pypeline.child_ancestors_incomplete "
                    "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                    "execution_id=%s task_name=%s child=%s",
                    pipeline_id,
                    pipeline_name,
                    actor_name,
                    message_id,
                    execution_id,
                    task_name,
                    child,
                )
                continue

            if base_case_execution_id == execution_id:
                for scenario in message.options["scenarios"]:
                    child_ancestors = list(graph.predecessors(child))
                    child_has_other_ancestors_in_scenario = False

                    for ancestor in child_ancestors:
                        if ancestor in scenario["tasksToRunInScenario"]:
                            child_has_other_ancestors_in_scenario = True
                            break

                    if (
                        child in scenario["tasksToRunInScenario"]
                        and task_name in child_ancestors
                        and task_name not in scenario["tasksToRunInScenario"]
                        and not child_has_other_ancestors_in_scenario
                    ):
                        scenario_dispatch_key = self._child_dispatched_key(
                            base_case_execution_id, scenario["execution_id"], child
                        )
                        if not self.redis_conn.safe_set(
                            scenario_dispatch_key, "1", nx=True, ex=DEFAULT_RESULT_TTL
                        ):
                            logger.info(
                                "pypeline.scenario_child_already_dispatched "
                                "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                                "execution_id=%s task_name=%s child=%s scenario_execution_id=%s",
                                pipeline_id,
                                pipeline_name,
                                actor_name,
                                message_id,
                                execution_id,
                                task_name,
                                child,
                                scenario["execution_id"],
                            )
                            continue

                        task_key = f"{scenario['execution_id']}-{child}"
                        locking_parallel_barrier = LockingParallelBarrier(
                            self.redis_url,
                            task_key=task_key,
                            lock_key=f"{base_case_execution_id}-lock",
                        )
                        locking_parallel_barrier.set_task_count(1)
                        scenario_repl = scenario.get("taskReplacements") or {}
                        handler = task_definitions[child]["handlers"][
                            scenario_repl.get(child, 0)
                        ]
                        server_type = task_definitions[child].get("serverType", None)

                        lazy_actor = register_lazy_actor(
                            broker,
                            get_callable(handler),
                            pipeline_config["metadata"],
                            server_type,
                        )
                        scenario_message = lazy_actor.message()
                        scenario_message.options["pipeline"] = pipeline
                        if max_retries is not None:
                            scenario_message.options["max_retries"] = max_retries
                        scenario_message.options["task_replacements"] = scenario_repl
                        scenario_message.options["execution_id"] = scenario[
                            "execution_id"
                        ]

                        scenario_message.options["task_name"] = child
                        scenario_message.options["base_case_execution_id"] = (
                            message.options["base_case_execution_id"]
                        )
                        scenario_message.options["scenarios"] = message.options[
                            "scenarios"
                        ]
                        if "settings" in message.kwargs:
                            scenario_message.kwargs["settings"] = copy(
                                message.kwargs["settings"]
                            )
                            scenario_message.kwargs["settings"]["execution_id"] = (
                                scenario["execution_id"]
                            )

                        logger.info(
                            "pypeline.enqueue_scenario_child "
                            "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                            "execution_id=%s task_name=%s child=%s scenario_execution_id=%s",
                            pipeline_id,
                            pipeline_name,
                            actor_name,
                            message_id,
                            execution_id,
                            task_name,
                            child,
                            scenario["execution_id"],
                        )
                        messages.append(scenario_message)

            child_dispatch_key = self._child_dispatched_key(
                base_case_execution_id, execution_id, child
            )
            if not self.redis_conn.safe_set(
                child_dispatch_key, "1", nx=True, ex=DEFAULT_RESULT_TTL
            ):
                logger.info(
                    "pypeline.child_already_dispatched "
                    "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                    "execution_id=%s task_name=%s child=%s",
                    pipeline_id,
                    pipeline_name,
                    actor_name,
                    message_id,
                    execution_id,
                    task_name,
                    child,
                )
                continue

            task_key = f"{execution_id}-{child}"
            locking_parallel_barrier = LockingParallelBarrier(
                self.redis_url,
                task_key=task_key,
                lock_key=f"{base_case_execution_id}-lock",
            )
            locking_parallel_barrier.set_task_count(1)
            handler = task_definitions[child]["handlers"][
                task_replacements.get(child, 0)
            ]
            server_type = task_definitions[child].get("serverType", None)
            lazy_actor = register_lazy_actor(
                broker,
                get_callable(handler),
                pipeline_config["metadata"],
                server_type,
            )

            child_message = lazy_actor.message()
            child_message.options["pipeline"] = pipeline
            if max_retries is not None:
                child_message.options["max_retries"] = max_retries
            child_message.options["task_replacements"] = task_replacements
            child_message.options["execution_id"] = execution_id
            child_message.options["task_name"] = child
            child_message.options["base_case_execution_id"] = message.options[
                "base_case_execution_id"
            ]
            child_message.options["scenarios"] = message.options["scenarios"]
            if "settings" in message.kwargs:
                child_message.kwargs["settings"] = message.kwargs["settings"]
                child_message.kwargs["settings"]["execution_id"] = message.options[
                    "execution_id"
                ]

            logger.info(
                "pypeline.enqueue_child "
                "pipeline_id=%s pipeline_name=%s actor_name=%s message_id=%s "
                "execution_id=%s task_name=%s child=%s",
                pipeline_id,
                pipeline_name,
                actor_name,
                message_id,
                execution_id,
                task_name,
                child,
            )
            messages.append(child_message)

        for new_message in messages:
            broker.enqueue(new_message)



import random
import time
from urllib.parse import urlparse

import redis
from redis.backoff import ExponentialBackoff
from redis.exceptions import (
    BusyLoadingError,
    ConnectionError,
    TimeoutError,
    TryAgainError,
)
from redis.retry import Retry
from redis.sentinel import MasterNotFoundError, Sentinel

from pypeline.constants import (
    REDIS_SENTINEL_MASTER_NAME,
    DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
    DEFAULT_REDIS_SOCKET_TIMEOUT,
    DEFAULT_REDIS_RETRY_ON_TIMEOUT,
    DEFAULT_REDIS_SOCKET_KEEPALIVE,
    DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
)


TRANSIENT_REDIS_ERRORS = (
    ConnectionError,
    TimeoutError,
    MasterNotFoundError,
    BusyLoadingError,
    TryAgainError,
)

DEFAULT_TIMEOUT = 180.0
DEFAULT_RETRY_INTERVAL = 0.25
MAX_BACKOFF = 5.0


class ResilientRedisClient:
    def __init__(self, redis_url, decode_responses=False):
        self.redis_url = redis_url
        self.redis = self._get_redis_connection(decode_responses=decode_responses)

    def _get_redis_connection(self, decode_responses=False):
        retry = Retry(ExponentialBackoff(), 5)

        common_kwargs = {
            "socket_connect_timeout": DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
            "socket_timeout": DEFAULT_REDIS_SOCKET_TIMEOUT,
            "retry_on_timeout": DEFAULT_REDIS_RETRY_ON_TIMEOUT,
            "socket_keepalive": DEFAULT_REDIS_SOCKET_KEEPALIVE,
            "health_check_interval": DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
            "retry": retry,
            "retry_on_error": [
                ConnectionError,
                TimeoutError,
                MasterNotFoundError,
                BusyLoadingError,
                TryAgainError,
            ],
        }

        if REDIS_SENTINEL_MASTER_NAME is not None:
            parsed_redis_url = urlparse(self.redis_url)
            redis_sentinel = Sentinel(
                sentinels=[(parsed_redis_url.hostname, parsed_redis_url.port)],
                sentinel_kwargs=common_kwargs,
            )
            return redis_sentinel.master_for(
                REDIS_SENTINEL_MASTER_NAME,
                db=int(parsed_redis_url.path[1:]) if parsed_redis_url.path else 0,
                password=parsed_redis_url.password,
                decode_responses=decode_responses,
                **common_kwargs,
            )

        return redis.StrictRedis.from_url(
            self.redis_url,
            decode_responses=decode_responses,
            **common_kwargs,
        )

    def _sleep_with_exponential_backoff(
        self, attempt: int, retry_interval: float
    ) -> None:
        delay = retry_interval * (2**attempt)
        delay = min(delay, 5.0)
        jitter = min(delay * 0.2, 1.0)
        delay += random.uniform(0, jitter)
        time.sleep(delay)

    def call(
        self,
        fn,
        *,
        timeout: float = DEFAULT_TIMEOUT,
        retry_interval: float = DEFAULT_RETRY_INTERVAL,
        op_name: str = "redis operation",
        retry_on_result=None,
    ):
        deadline = time.monotonic() + timeout
        attempt = 0
        last_error = None

        while True:
            try:
                result = fn()

                if retry_on_result is not None and retry_on_result(result):
                    if time.monotonic() >= deadline:
                        raise TimeoutError(
                            f"{op_name} timed out after {timeout}s without success"
                        )
                    self._sleep_with_exponential_backoff(attempt, retry_interval)
                    attempt += 1
                    continue

                return result

            except TRANSIENT_REDIS_ERRORS as exc:
                last_error = exc
                if time.monotonic() >= deadline:
                    raise TimeoutError(
                        f"{op_name} failed after {timeout}s; last error: {exc}"
                    ) from exc

                self._sleep_with_exponential_backoff(attempt, retry_interval)
                attempt += 1

    def safe_set(
        self,
        *args,
        timeout=DEFAULT_TIMEOUT,
        retry_interval=DEFAULT_RETRY_INTERVAL,
        retry_on_result=None,
        **kwargs,
    ):
        return self.call(
            lambda: self.redis.set(*args, **kwargs),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name="redis.set",
            retry_on_result=retry_on_result,
        )

    def safe_exists(
        self,
        *args,
        timeout=DEFAULT_TIMEOUT,
        retry_interval=DEFAULT_RETRY_INTERVAL,
        **kwargs,
    ):
        return self.call(
            lambda: self.redis.exists(*args, **kwargs),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name="redis.exists",
        )

    def safe_delete(
        self,
        *args,
        timeout=DEFAULT_TIMEOUT,
        retry_interval=DEFAULT_RETRY_INTERVAL,
        **kwargs,
    ):
        return self.call(
            lambda: self.redis.delete(*args, **kwargs),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name="redis.delete",
        )

    def safe_get(
        self,
        *args,
        timeout=DEFAULT_TIMEOUT,
        retry_interval=DEFAULT_RETRY_INTERVAL,
        **kwargs,
    ):
        return self.call(
            lambda: self.redis.get(*args, **kwargs),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name="redis.get",
        )

    def safe_setex(
        self,
        *args,
        timeout=DEFAULT_TIMEOUT,
        retry_interval=DEFAULT_RETRY_INTERVAL,
        **kwargs,
    ):
        return self.call(
            lambda: self.redis.setex(*args, **kwargs),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name="redis.setex",
        )



import contextvars
from dramatiq import Middleware


class GetActiveWorkerIdMiddleware(Middleware):
    _ACTIVE_WORKER_ID: contextvars.ContextVar["Optional[Message[Any]]"] = (
        contextvars.ContextVar("_ACTIVE_WORKER_ID", default=None)
    )

    @classmethod
    def get_active_worker_id(cls):
        return cls._ACTIVE_WORKER_ID.get()

    def before_process_message(self, broker, message):
        """Sets the active worker ID, verifying broker has a broker_id first."""
        broker_id = getattr(broker, "broker_id", None)
        if not broker_id:
            return
        self._ACTIVE_WORKER_ID.set(broker_id)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        self._ACTIVE_WORKER_ID.set(None)



from __future__ import annotations

import copy
import json
import time
import typing
from uuid import uuid4

from dramatiq.broker import get_broker
from dramatiq.results import ResultMissing
from redis.exceptions import RedisError

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import (
    REDIS_URL,
    DEFAULT_RESULT_TTL,
)
from pypeline.utils.redis_client import ResilientRedisClient

from dramatiq.message import Message


class parallel_pipeline:
    """Chain actors together, passing the result of one actor to the
    next one in line.

    Parameters:
      children(typing.List[typing.List[Message]]): A sequence of messages or
        pipelines. Child pipelines are flattened into the resulting
        pipeline.
      broker(Broker): The broker to run the pipeline on. Defaults to
        the current global broker.
    """

    messages: list[Message]

    def __init__(self, messages: typing.List[typing.List[Message]], broker=None):
        self.broker = broker or get_broker()
        self.messages = messages
        self.execution_id = str(uuid4())
        execution_graph = []

        for group_index, message_group in enumerate(self.messages):
            sub_execution_group = []
            group_completion_uuid = str(uuid4())
            for task_index, m in enumerate(message_group):
                m.kwargs["event"]["execution_id"] = self.execution_id
                m.options["group_completion_uuid"] = group_completion_uuid
                m.options["group_index"] = group_index
                m.options["task_index"] = task_index
                message_dict = copy.deepcopy(m.asdict())
                sub_execution_group.append(message_dict)
            execution_graph.append(sub_execution_group)

        self.execution_graph = execution_graph

        for m in self.messages[0]:
            m.options["execution_graph"] = execution_graph

    def __len__(self):
        """Returns the length of the parallel_pipeline."""
        count = 0
        for message_group in self.messages:
            count += len(message_group)
        return count

    def __str__(self):  # pragma: no cover
        """Return a string representation of the parallel_pipeline."""
        result = []

        for i, message_group in enumerate(self.messages):
            group_str = f"Group {i + 1}: [\n"
            for j, message in enumerate(message_group):
                group_str += f"  Message {j + 1}: {message.actor_name}\n"
            group_str += "]\n"
            result.append(group_str)

        return "".join(result)

    @property
    def completed(self):
        return self.completed_count == len(self)

    @property
    def completed_count(self):
        count = 0

        for message_group in self.messages:
            for message in message_group:
                try:
                    message.get_result()
                    count += 1
                except ResultMissing:
                    pass
        return count

    def run(self, *, delay=None):
        """Run this parallel_pipeline."""
        starting_group = self.messages[0]

        completion_uuid = starting_group[0].options["group_completion_uuid"]
        locking_parallel_barrier = LockingParallelBarrier(
            REDIS_URL, task_key=completion_uuid, lock_key=f"{completion_uuid}-lock"
        )
        locking_parallel_barrier.set_task_count(len(starting_group))

        redis_conn = ResilientRedisClient(REDIS_URL)
        redis_conn.safe_set(
            self._group_dispatched_key(self.execution_id, 0),
            "1",
            ex=DEFAULT_RESULT_TTL,
        )

        for m in starting_group:
            self.broker.enqueue(m, delay=delay)

        return self

    def get_result(self, *, block=False, timeout=None):
        """Get the result of this pipeline."""
        last_message = self.messages[-1][-1]
        backend = self.broker.get_results_backend()
        return last_message.get_result(backend=backend, block=block, timeout=timeout)

    def get_results(self, *, block=False, timeout=None):
        """Get the results of each job in the pipeline."""
        deadline = None
        if timeout:
            deadline = time.monotonic() + timeout / 1000

        for message_group in self.messages:
            for message in message_group:
                if deadline:
                    timeout = max(0, int((deadline - time.monotonic()) * 1000))

                backend = self.broker.get_results_backend()
                yield {
                    message.actor_name: message.get_result(
                        backend=backend, block=block, timeout=timeout
                    )
                }

    def to_json(self) -> str:
        """Convert the execution graph to a JSON string representation."""
        return json.dumps(self.execution_graph)

    @classmethod
    def from_json(cls, json_data: str) -> parallel_pipeline:
        """Reconstruct a parallel_pipeline from serialized execution graph."""
        execution_graph = json.loads(json_data)

        obj = cls.__new__(cls)
        obj.broker = get_broker()
        obj.execution_graph = execution_graph
        obj.messages = [
            [Message(**message_dict) for message_dict in message_group]
            for message_group in execution_graph
        ]

        try:
            obj.execution_id = obj.messages[0][0].kwargs["event"]["execution_id"]
        except (IndexError, KeyError):
            obj.execution_id = None

        return obj

    def _group_dispatched_key(self, execution_id: str, group_index: int) -> str:
        return f"pp:{execution_id}:group:{group_index}:dispatched"


class PipelineResult:
    """
    A class to manage and retrieve the results of a parallel pipeline execution.
    """

    def __init__(self, execution_id: str, result_ttl: int = DEFAULT_RESULT_TTL):
        self.pipeline: parallel_pipeline = None
        self.execution_id = execution_id
        self.redis_key = f"{execution_id}-results-key"
        self.result_ttl = result_ttl
        self.redis_conn = ResilientRedisClient(REDIS_URL)

    def create_result_entry(self, pipeline_json_str: str):
        if not pipeline_json_str:
            raise ValueError("No pipeline data passed to create result store")

        try:
            self.redis_conn.safe_setex(
                self.redis_key, self.result_ttl, pipeline_json_str
            )
        except RedisError as e:
            raise RuntimeError(f"Failed to store pipeline data in Redis: {e}")

    def load(self):
        try:
            pipeline_data = self.redis_conn.safe_get(self.redis_key)
            if pipeline_data:
                if isinstance(pipeline_data, bytes):
                    pipeline_data = pipeline_data.decode("utf-8")
                self.pipeline = parallel_pipeline.from_json(pipeline_data)
            else:
                self.pipeline = None
        except RedisError as e:
            raise RuntimeError(f"Failed to load pipeline data from Redis: {e}")

    @property
    def status(self) -> str:
        if not self.pipeline:
            return "unavailable"
        return "complete" if self.pipeline.completed else "pending"

    def get_results(self) -> dict:
        if not self.pipeline:
            return {}

        results = {}
        for result in self.pipeline.get_results():
            for actor, res in result.items():
                unique_actor = self._get_unique_actor_name(actor, results)
                results[unique_actor] = res
        return results

    def get_result(self):
        if self.pipeline:
            return self.pipeline.get_result()

    def _get_unique_actor_name(self, actor: str, results: dict) -> str:
        if actor not in results:
            return actor

        suffix = 0
        new_actor = f"{actor}-{suffix}"
        while new_actor in results:
            suffix += 1
            new_actor = f"{actor}-{suffix}"
        return new_actor



"""PypeLine Library Setup"""

import re
import ast
import sys
import os
from typing import List

from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """PyTest Command"""

    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import coverage
        import pytest

        if self.pytest_args and len(self.pytest_args) > 0:
            self.test_args.extend(self.pytest_args.strip().split(" "))
            self.test_args.append("tests/")

        cov = coverage.Coverage()
        cov.start()
        errno = pytest.main(self.test_args)
        cov.stop()
        cov.report()
        cov.html_report()
        print("Wrote coverage report to htmlcov directory")
        sys.exit(errno)


# Defaults that will be overridden/updated if do_cythonize is true.
#
ext_modules = []
cmdclass = {"test": PyTest}
packages = find_packages(exclude=["tests"])

do_cythonize = os.getenv("CYTHONIZE", "false").lower() == "true"
if do_cythonize:
    try:
        from Cython.Build import cythonize
        from Cython.Distutils import build_ext

        cmdclass["build_ext"] = build_ext  # Use cython's build_ext

        def scandir(
            directory: str,
            exclude_paths: List[str] = None,
            paths: List[str] = None,
            directories: List[str] = None,
            return_directories: bool = False,
            recursive: bool = True,
        ):
            """Scan a given directory recursively and produce a list
            of file paths (e.g. ['pypeline/app.py', 'pypeline/config.py'])
            or directroies (e.g. ['pypeline/api', 'pypeline/utils'])
            """
            if paths is None:
                paths = []
            if directories is None:
                directories = []
            if exclude_paths is None:
                exclude_paths = []

            for file in os.listdir(directory):
                path = os.path.join(directory, file)
                if any([p in path for p in exclude_paths]):
                    continue

                if "__init__.py" in path or "__pycache__" in path:
                    continue

                if os.path.isfile(path) and path.endswith(".py"):
                    paths.append(path)
                elif os.path.isdir(path) and recursive:
                    directories.append(path)
                    scandir(
                        path,
                        exclude_paths=exclude_paths,
                        paths=paths,
                        directories=directories,
                        return_directories=return_directories,
                        recursive=recursive,
                    )

            if return_directories:
                return directories
            return paths

        def make_extension(ext_path: str):
            """Generate an Extension() object. Takes a path
            e.g. pypeline/app.py
            and generates a valid Extension
            e.g. Extension('pypeline.app', ['pypeline/app.py'])
            """
            extName = ext_path.replace("/", ".")[:-3]
            return Extension(extName, [ext_path], include_dirs=["."])

        to_cythonize = scandir(
            "pypeline",
            exclude_paths=[
                "pypeline/templates",
                "pypeline/static",
                "pypeline/celery.py",
                "pypeline/tools/thumbnail/thumbnail.py",
                "pypeline/lib/config_server.py",
            ],
        )

        ext_modules = cythonize(
            [make_extension(path) for path in to_cythonize], language_level=3
        )

        class MyBuildPy(build_py):
            """Update standard build_py to exclude any files we
            explicitly cythonize.
            """

            def find_package_modules(self, package, package_dir):
                """Return list of all package modules that are *not* cythonized
                for regular packaging.
                """
                modules = super().find_package_modules(package, package_dir)
                filtered_modules = []
                for tup in modules:
                    if tup[2] in to_cythonize:
                        continue
                    filtered_modules.append(tup)
                return filtered_modules

        cmdclass["build_py"] = MyBuildPy

    except ImportError:
        pass

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("pypeline/__init__.py", "rb") as f:
    __version__ = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

setup(
    name="scalable_pypeline",
    version=__version__,
    description="PypeLine - Python pipelines for the Real World",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="Bravos Power Corporation",
    license="Apache License 2.0",
    url="https://gitlab.com/bravos2/pypeline",
    packages=packages,
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    install_requires=install_requires,
    extras_require={
        "build": ["wheel", "twine"],
        "flask": [
            "markupsafe==2.0.1",
            "flask-smorest>=0.23.0,<1",
        ],
        "web": ["gunicorn", "gevent>=24.10.1,<26"],
        "workers": [
            "networkx>=2.4",
            "dramatiq[rabbitmq]==1.17.0",
            "apscheduler>=3.10.4,<4",
            "tenacity==8.0.1",
        ],
        "dev": ["black"],
        "test": [
            "pytest-cov>=2.6.1,<3",
            "tox>=3.14.1,<4",
            "mock>=1,<2",
            "responses>=0.10.16,<0.11",
            "fakeredis>=2.10.3,<2.31",  # fakeredis version compatible with redis 4.x
        ],
    },
    entry_points={
        "flask.commands": [
            "pypeline-worker=pypeline.dramatiq:pypeline_worker",
            "cron-scheduler=pypeline.dramatiq:cron_scheduler",
        ],
        "console_scripts": ["job-runner = pypeline.job_runner:main"],
    },
)



import uuid

from redis.exceptions import (
    BusyLoadingError,
    ConnectionError,
    TimeoutError,
    TryAgainError,
)
from redis.sentinel import MasterNotFoundError

from pypeline.utils.redis_client import ResilientRedisClient


TRANSIENT_REDIS_ERRORS = (
    ConnectionError,
    TimeoutError,
    MasterNotFoundError,
    BusyLoadingError,
    TryAgainError,
)


class LockingParallelBarrier:
    def __init__(self, redis_url, task_key="task_counter", lock_key="task_lock"):
        self.redis_client = ResilientRedisClient(redis_url, decode_responses=True)
        self.redis = self.redis_client.redis

        self.task_key = task_key
        self.lock_key = lock_key
        self.lock_token = None

        self._release_lock_script = self.redis.register_script(
            """
            if redis.call("get", KEYS[1]) == ARGV[1] then
                return redis.call("del", KEYS[1])
            else
                return 0
            end
            """
        )

    def _sleep_with_exponential_backoff(
        self, attempt: int, retry_interval: float
    ) -> None:
        self.redis_client._sleep_with_exponential_backoff(attempt, retry_interval)

    def _call_with_failover_retry(
        self,
        fn,
        *,
        timeout: float = 180.0,
        retry_interval: float = 0.25,
        op_name: str = "redis operation",
        retry_on_result=None,
    ):
        deadline = __import__("time").monotonic() + timeout
        attempt = 0

        while True:
            try:
                result = fn()

                if retry_on_result is not None and retry_on_result(result):
                    if __import__("time").monotonic() >= deadline:
                        raise TimeoutError(
                            f"{op_name} timed out after {timeout}s without success"
                        )
                    self._sleep_with_exponential_backoff(attempt, retry_interval)
                    attempt += 1
                    continue

                return result

            except TRANSIENT_REDIS_ERRORS as exc:
                if __import__("time").monotonic() >= deadline:
                    raise TimeoutError(
                        f"{op_name} failed after {timeout}s; last error: {exc}"
                    ) from exc

                self._sleep_with_exponential_backoff(attempt, retry_interval)
                attempt += 1

    def advance_once(self, advance_key: str, ttl: int) -> bool:
        """
        Returns True exactly once per advance_key (within ttl).
        Subsequent calls return False.
        """
        return bool(
            self._call_with_failover_retry(
                lambda: self.redis.set(advance_key, "1", nx=True, ex=ttl),
                op_name=f"advance_once({advance_key})",
            )
        )

    def acquire_lock(self, timeout=180, retry_interval=0.25):
        """
        Acquire a lock using Redis and store an ownership token.

        Preserves the existing public signature:
        - timeout: total wait budget AND lock TTL
        - retry_interval: base interval for exponential backoff

        Raises TimeoutError if the lock cannot be acquired within timeout seconds.
        """
        token = str(uuid.uuid4())

        result = self._call_with_failover_retry(
            lambda: self.redis.set(self.lock_key, token, nx=True, ex=timeout),
            timeout=timeout,
            retry_interval=retry_interval,
            op_name=f"acquire_lock({self.lock_key})",
            retry_on_result=lambda result: not bool(result),
        )

        if result:
            self.lock_token = token
            return True

        raise TimeoutError(
            f"Could not acquire lock '{self.lock_key}' within {timeout}s"
        )

    def release_lock(self):
        """Release the lock only if this instance still owns it."""
        if not self.lock_token:
            return False

        token = self.lock_token
        try:
            released = self._call_with_failover_retry(
                lambda: self._release_lock_script(
                    keys=[self.lock_key],
                    args=[token],
                ),
                op_name=f"release_lock({self.lock_key})",
            )
            return bool(released)
        finally:
            self.lock_token = None

    def set_task_count(self, count):
        """Initialize the task counter in Redis."""
        self._call_with_failover_retry(
            lambda: self.redis.set(self.task_key, count),
            op_name=f"set_task_count({self.task_key})",
        )

    def decrement_task_count(self):
        """Decrement the task counter in Redis."""
        return int(
            self._call_with_failover_retry(
                lambda: self.redis.decr(self.task_key),
                op_name=f"decrement_task_count({self.task_key})",
            )
        )

    def task_exists(self):
        return bool(
            self._call_with_failover_retry(
                lambda: self.redis.exists(self.task_key),
                op_name=f"task_exists({self.task_key})",
            )
        )

    def get_task_count(self):
        """Get the current value of the task counter."""
        value = self._call_with_failover_retry(
            lambda: self.redis.get(self.task_key),
            op_name=f"get_task_count({self.task_key})",
        )
        return int(value or 0)
